// Copyright 2024 OpenPubkey
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// SPDX-License-Identifier: Apache-2.0

package gq

import (
	"math/big"

	"filippo.io/bigmod"
)

// leaks only the size of x
func modAsInt(x *bigmod.Modulus) *big.Int {
	return new(big.Int).SetBytes(x.Nat().Bytes(x))
}

// leaks only the size of x
func natAsInt(x *bigmod.Nat, m *bigmod.Modulus) *big.Int {
	return new(big.Int).SetBytes(x.Bytes(m))
}

// leaks only the size of x
func intAsNat(x *big.Int, m *bigmod.Modulus) (*bigmod.Nat, error) {
	return bigmod.NewNat().SetBytes(x.Bytes(), m)
}
